/***************************************************************************
	mmquicktime - Classes for creating Quicktime (.mov) files

			-------------------
        begin           : 6/4/2004
        copyright       : (C) 2004 by Michael Minn
        email           : see www.michaelminn.com
                        -------------------

    This program is free software; you can redistribute it and/or modify
    it under the terms of version 2 of the GNU General Public License as 
    published by the Free Software Foundation.

    Y = (0.299 * r) + (0.587 * g) + (0.114 * b)
    Cr = U = (r - y) / 1.402
    Cb = V = (b - y) / 1.772

    r = y + (1.402 * u)
    g = y - (0.344 * v) - (0.714 * u)
    b = y + (1.772 * v)

    // Apple
    r = 1.402 * v + y + 0.5
    g = y - (0.7143 * v) - (0.3437 * u) + .5
    b = (1.77 * u) + y + .5
***************************************************************************/

//#include <stdio.h>
//#include <stdlib.h>
//#include <string.h>
#include <errno.h>
#include <time.h>
//#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include "mmquicktime.h"

unsigned int little_endian(unsigned char *big_endian_value)
{
	return (((unsigned int) big_endian_value[0]) << 24)
	     + (((unsigned int) big_endian_value[1]) << 16)
	     + (((unsigned int) big_endian_value[2]) << 8)
	     + ((unsigned int) big_endian_value[3]);
}

unsigned int little_endian_short(unsigned char *big_endian_value)
{
	short value = (((short) big_endian_value[0]) << 8) + ((short) big_endian_value[1]);
	return value;
}

int big_endian(unsigned int little_endian, unsigned char *big_endian)
{
	big_endian[0] = (little_endian >> 24) & 0xff;
	big_endian[1] = (little_endian >> 16) & 0xff;
	big_endian[2] = (little_endian >> 8) & 0xff;
	big_endian[3] = little_endian & 0xff;
	return *((int*) big_endian);
}

int big_endian_short(unsigned int little_endian, unsigned char *big_endian)
{
	big_endian[0] = (little_endian >> 8) & 0xff;
	big_endian[1] = little_endian & 0xff;
	return *((int*) big_endian);
}

int mmquicktime_list(char *filename)
{
	int descriptor = -1;
	unsigned char *buffer;
	struct stat file_info;
	
	if ((descriptor = open(filename, O_RDONLY)) < 0)
		perror(filename);
			
	else if (fstat(descriptor, &file_info) < 0)
		perror(filename);
			
	else if ((buffer = new unsigned char[file_info.st_size]) == NULL)
		perror("Buffer allocation failure");

	else if (read(descriptor, buffer, file_info.st_size) < 0)
		perror(filename);
		
	else
	{
		unsigned char *scan = buffer; 
		while (scan < &buffer[file_info.st_size])
		{
			//int is_container = 0;
			char *atom_name = (char*) &scan[4];

			unsigned int atom_size = little_endian(scan);
			unsigned int actual_size = atom_size;
			if ((strncmp(atom_name, "dref", 4) == 0) || (strncmp(atom_name, "stsd", 4) == 0))
				actual_size = 16; // entries in data ref & sample desc table are atoms

			if (atom_size == 1) // strange kludge for file created with libquicktime
				actual_size = little_endian(&scan[12]);

			if ((strncmp(atom_name, "moov", 4) == 0)
			    || (strncmp(atom_name, "trak", 4) == 0)
			    || (strncmp(atom_name, "mdia", 4) == 0)
			    || (strncmp(atom_name, "clip", 4) == 0)
			    || (strncmp(atom_name, "matt", 4) == 0)
			    || (strncmp(atom_name, "edts", 4) == 0)
			    || (strncmp(atom_name, "dinf", 4) == 0)
			    || (strncmp(atom_name, "minf", 4) == 0)
			    || (strncmp(atom_name, "udta", 4) == 0)
			    || (strncmp(atom_name, "stbl", 4) == 0))
				actual_size = 8;
				// is_container = 1;

			// dinf atom names start with a9

			if (actual_size == 8) // container
				printf("Atom %.4s = %d bytes @ %x\n", 
					atom_name, atom_size, scan - buffer);
			else
				printf("  Atom %.4s = %d (%d) bytes @ %x, version %x, flags %02x%02x%02x\n", 
					atom_name, atom_size, actual_size, scan - buffer, 
					scan[8], scan[9], scan[10], scan[11]);
	
			if (strncmp(atom_name, "tkhd", 4) == 0)
				((mmquicktime_track_header*) scan)->list();

			else if (strncmp(atom_name, "mvhd", 4) == 0)
				((mmquicktime_movie_header*) scan)->list();

			else if (strncmp(atom_name, "mdhd", 4) == 0)
				((mmquicktime_media_header*) scan)->list();

			else if (strncmp(atom_name, "hdlr", 4) == 0)
				((mmquicktime_media_handler_reference*) scan)->list();

			else if (strncmp(atom_name, "vmhd", 4) == 0)
				((mmquicktime_video_media_information*) scan)->list();

			else if (strncmp(atom_name, "dref", 4) == 0)
				((mmquicktime_data_reference*) scan)->list();

			else if (strncmp(atom_name, "stsd", 4) == 0)
				((mmquicktime_sample_description*) scan)->list();

			else if (strncmp(atom_name, "stts", 4) == 0)
				((mmquicktime_time_to_sample_table*) scan)->list();

			else if (strncmp(atom_name, "stss", 4) == 0)
				((mmquicktime_keyframe_table*) scan)->list();

			else if (strncmp(atom_name, "stsc", 4) == 0)
				((mmquicktime_chunk_table*) scan)->list();

			else if (strncmp(atom_name, "stsz", 4) == 0)
				((mmquicktime_sample_size_table*) scan)->list();

			else if (strncmp(atom_name, "stco", 4) == 0)
				((mmquicktime_sample_offset_table*) scan)->list();

			else if (strncmp(atom_name, "ctab", 4) == 0)
				((mmquicktime_color_table*) scan)->list();

			if (actual_size <= 0)
				scan = &buffer[file_info.st_size];
			else
 				scan += actual_size;
		}
	}
	
	if (descriptor >= 0)
		close(descriptor);
		
	if (buffer)
		delete buffer;
	return 0;
}

char *mmquicktime_date_string(unsigned int quicktime_time)
{
	// quicktime_time is seconds since 1/1/1904
	int four_year = quicktime_time / 126230400;
	int four_year_seconds = quicktime_time % 126230400;

	int year = 1904 + (4 * four_year);
	int year_seconds = four_year_seconds;
	if (four_year_seconds >= 31622400)
	{
		year += (1 + ((four_year_seconds - 31622400) / 31536000));
		year_seconds = ((four_year_seconds - 31622400) % 31536000);
	}

	int year_day = year_seconds / (24 * 60 * 60);
	int hour = (year_seconds % (24 * 60 * 60)) / (60 * 60);
	int minute = (year_seconds % (60 * 60)) / 60;

	int day = 0;
	int month = 0;
	int month_length[12] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
	for (int scan = 0, sum = 0; sum < year_day; ++scan)
	{
		int length = month_length[scan];
		if ((month == 1) && ((year % 4) == 0)) // leap year
			++length;
		if ((sum < year_day) && (year_day < (sum + length)))
		{
			month = scan + 1;
			day = year_day - sum + 1;
		}
		sum += length;
	}

	static char date_string[32];
	sprintf(date_string, "%02d/%02d/%04d %02d:%02d", month, day, year, hour, minute);
	return date_string;
}


/*********************************************************/

int mmquicktime_file::open_file(char *filename, int color_mode, int frames_per_second)
{
	if (!this)
		return -EINVAL;

	if (frames_per_second <= 0)
	{
		fprintf(stderr, "Invalid frames per second value %d\n", frames_per_second);
		return -EINVAL;
	}

	memset(this, 0, sizeof(*this));

	this->color_mode = color_mode;
	this->frames_per_second = frames_per_second;

	if ((descriptor = open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0664)) < 0)
	{
		int status = -errno;
		perror(filename);
		return status;
	}

	unsigned char header[16] = { 0, 0, 0, 0, 'm', 'd', 'a', 't' };

	if (write(descriptor, header, 8) < 0)
	{
		int status = -errno;
		perror(filename);
		close(descriptor);
		descriptor = -1;
		return status;
	}

	return descriptor;	
}

int mmquicktime_file::add_frame(mmpnm_file *frame)
{
	if (!this || (descriptor < 0))
		return -EINVAL;

	if (!frame_width || !frame_height)
	{
		frame_width = frame->width;
		frame_height = frame->height;
	}
	
	else if ((frame_width != frame->width) || (frame_height != frame->height))
	{
		fprintf(stderr, "Cropping %d x %d frame to %d x %d\n", 
			frame->width, frame->height, frame_width, frame_height);
		frame->crop(frame_width, frame_height);
	}

	int length = frame_width * frame_height * 3;
	unsigned char *data = frame->color_data;
	if (color_mode == MMQUICKTIME_COLOR_MODE_GRAY)
	{
		data = frame->gray_data;
		length = frame_width * frame_height;
	}

	if (write(descriptor, data, length) < 0)
	{
		int status = -errno;
		perror("Failure writing frame data");
		return status;
	}

	++frame_count;

	return length;
}

int mmquicktime_file::close_file()
{
	int length_sum = 0;
	unsigned int creation_time = MMQUICKTIME_TIME_OFFSET_SECONDS + time(NULL);
	unsigned char default_matrix[36] = 
		{ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		  0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0,
		  0, 0, 0, 0, 0, 0, 0, 0, 0x40, 0, 0, 0 };

	// Find length of movie data
	struct stat file_info;
	if (fstat(descriptor, &file_info) < 0)
	{
		int status = -errno;
		perror("Failure finding movie data length");
		return status;
	}

	int movie_data_length = file_info.st_size;
	int time_per_frame = MMQUICKTIME_TIME_SCALE / frames_per_second;
	int duration = frame_count * time_per_frame;

	int frame_size = frame_width * frame_height * 3;
	if (color_mode == MMQUICKTIME_COLOR_MODE_GRAY)
		frame_size = frame_width * frame_height;

	// Create movie information atoms at end of file
	mmquicktime_container movie_container;
	memcpy(movie_container.atom_name, "moov", 4);
	length_sum += 8;

		mmquicktime_movie_header movie_header;
		memset(&movie_header, 0, sizeof(movie_header));
		big_endian(sizeof(movie_header), movie_header.atom_size);
		big_endian(duration, movie_header.duration);
		memcpy(movie_header.atom_name, "mvhd", 4);
		movie_header.version = 0;
		big_endian(creation_time, movie_header.creation_time);
		big_endian(creation_time, movie_header.modification_time);
		big_endian(MMQUICKTIME_TIME_SCALE, movie_header.time_scale);
		movie_header.preferred_rate[1] = 1;
		memset(movie_header.preferred_volume, 0, 2);
		memcpy(movie_header.matrix, default_matrix, 36);
		movie_header.next_track_id[3] = 2;
		length_sum += sizeof(movie_header);

		mmquicktime_container track_container;
		memcpy(track_container.atom_name, "trak", 4);
		int track_container_start = length_sum;
		length_sum += 8;

			mmquicktime_track_header track_header;
			memset(&track_header, 0, sizeof(track_header));
			big_endian(sizeof(track_header), track_header.atom_size);
			big_endian(duration, track_header.duration);
			memcpy(track_header.atom_name, "tkhd", 4);
			track_header.flags[2] = 0x0f; // enabled - [3] here caused no display
			big_endian(creation_time, track_header.creation_time);
			big_endian(creation_time, track_header.modification_time);
			big_endian(1, track_header.track_id);
			memcpy(track_header.matrix, default_matrix, 36);
			big_endian(frame_width << 16, track_header.track_width);
			big_endian(frame_height << 16, track_header.track_height);
			length_sum += sizeof(track_header);

			mmquicktime_container media_container;
			memcpy(media_container.atom_name, "mdia", 4);
			int media_container_start = length_sum;
			length_sum += 8;

				mmquicktime_media_header media_header;
				memset(&media_header, 0, sizeof(media_header));
				big_endian(sizeof(media_header), media_header.atom_size);
				big_endian(duration, media_header.duration);
				memcpy(media_header.atom_name, "mdhd", 4);
				big_endian(creation_time, media_header.creation_time);
				big_endian(creation_time, media_header.modification_time);
				big_endian(MMQUICKTIME_TIME_SCALE, media_header.time_scale);
				length_sum += sizeof(media_header);

				mmquicktime_media_handler_reference video_media_handler;
				memset(&video_media_handler, 0, sizeof(video_media_handler));
				big_endian(sizeof(video_media_handler), video_media_handler.atom_size);
				memcpy(video_media_handler.atom_name, "hdlr", 4);
				memcpy(video_media_handler.component_type, "mhlr", 4);
				memcpy(video_media_handler.component_subtype, "vide", 4);
				length_sum += sizeof(video_media_handler);

				/**
				memcpy(video_media_handler.component_manufacturer, "appl", 4);
				char *video_handler_name = "Apple Video Media Handler";
				int video_handler_length = sizeof(video_media_handler) 
						+ strlen(video_handler_name) + 1;
				big_endian(video_handler_length, video_media_handler.atom_size);
				length_sum += video_handler_length;
				**/

				mmquicktime_container media_information_container;
				memcpy(media_information_container.atom_name, "minf", 4);
				int media_information_container_start = length_sum;
				length_sum += 8;

					mmquicktime_video_media_information video_media_information;
					memset(&video_media_information, 0, 
						sizeof(video_media_information));
					big_endian(sizeof(video_media_information), 
						video_media_information.atom_size);
					memcpy(video_media_information.atom_name, "vmhd", 4);
					video_media_information.flags[2] = 1;
					video_media_information.graphics_mode[0] = 0;
					video_media_information.graphics_mode[1] = 0x40;
					video_media_information.transfer_mode_red[0] = 0x80;
					video_media_information.transfer_mode_green[0] = 0x80;
					video_media_information.transfer_mode_blue[0] = 0x80;
					length_sum += sizeof(video_media_information);

					mmquicktime_media_handler_reference video_alias_handler;
					memset(&video_alias_handler, 0, sizeof(video_alias_handler));
					big_endian(sizeof(video_alias_handler), 
						video_alias_handler.atom_size);
					memcpy(video_alias_handler.atom_name, "hdlr", 4);
					memcpy(video_alias_handler.component_type, "dhlr", 4);
					memcpy(video_alias_handler.component_subtype, "alis", 4);
					length_sum += sizeof(video_alias_handler);

					/**
					memcpy(video_alias_handler.component_manufacturer, "appl", 4);
					char *alias_handler_name = "Apple Alias Data Handler";
					int alias_handler_length = sizeof(video_media_handler) 
							+ strlen(alias_handler_name) + 1;
					big_endian(alias_handler_length, video_alias_handler.atom_size);
					length_sum += alias_handler_length;
					**/

					mmquicktime_container data_reference_container;
					memcpy(data_reference_container.atom_name, "dinf", 4);
					int data_reference_container_start = length_sum;
					length_sum += 8;

						mmquicktime_data_reference data_reference;
						memset(&data_reference, 0, sizeof(data_reference));
						big_endian(28, data_reference.atom_size);
						memcpy(data_reference.atom_name, "dref", 4);
						big_endian(1, data_reference.entry_count);
						length_sum += sizeof(data_reference);

						mmquicktime_data_reference_entry data_reference_entry;
						memset(&data_reference_entry, 0, 
							sizeof(data_reference_entry));
						memcpy(data_reference_entry.atom_name, "alis", 4);
						big_endian(12, data_reference_entry.atom_size);
						data_reference_entry.flags[2] = 1;
						length_sum += sizeof(data_reference_entry);

					big_endian(length_sum - data_reference_container_start,
						data_reference_container.atom_size);

					mmquicktime_container sample_table;
					memcpy(sample_table.atom_name, "stbl", 4);
					int sample_table_start = length_sum;
					length_sum += 8;

						mmquicktime_sample_description sample_description;
						memset(&sample_description, 0, sizeof(sample_description));
						memcpy(sample_description.atom_name, "stsd", 4);
						big_endian(16 + sizeof(mmquicktime_sample_description_entry), 
							sample_description.atom_size);
						big_endian(1, sample_description.entry_count);
						length_sum += sizeof(sample_description);

						mmquicktime_sample_description_entry sample_description_entry;
						memset(&sample_description_entry, 0, 
							sizeof(sample_description_entry));
						big_endian(sizeof(sample_description_entry), 
							sample_description_entry.atom_size);
						memcpy(sample_description_entry.atom_name, "raw ", 4);
						big_endian_short(24, sample_description_entry.pixel_depth);
						strcpy((char*)sample_description_entry.compression_algorithm_name,
							"\x10Uncompressed RGB");
						if (color_mode == MMQUICKTIME_COLOR_MODE_GRAY)
						{
							memcpy(sample_description_entry.atom_name, "raw ", 4);
							big_endian_short(8, sample_description_entry.pixel_depth);
						}

						sample_description_entry.reference_index[1] = 1;
						// Max quality:  from Inside Macintosh: Quicktime pp 3-455
						big_endian(0x3ff, sample_description_entry.temporal_quality);
						big_endian(0x3ff, sample_description_entry.spacial_quality);
						big_endian_short(frame_width, 
							sample_description_entry.width);
						big_endian_short(frame_height, 
							sample_description_entry.height);
						big_endian(72 << 16, 
							sample_description_entry.horizontal_resolution);
						big_endian(72 << 16, 
							sample_description_entry.vertical_resolution);
						big_endian_short(1, sample_description_entry.frame_count);
						memset(sample_description_entry.color_table_id, 0xff, 2);
						length_sum += sizeof(sample_description_entry);

						mmquicktime_color_table color_table;
						if (color_mode == MMQUICKTIME_COLOR_MODE_GRAY)
						{
							big_endian(sizeof(sample_description)
								+ sizeof(sample_description_entry)
								+ sizeof(color_table),
								sample_description.atom_size);
					
							big_endian(sizeof(sample_description_entry)
								+ sizeof(color_table),
								sample_description_entry.atom_size);

							big_endian_short(8, sample_description_entry.pixel_depth);
							sample_description_entry.color_table_id[0] = 0;
							sample_description_entry.color_table_id[1] = 0;

							memset(&color_table, 0, sizeof(color_table));
							color_table.flags[0] = 0x80;
							color_table.entry_count[1] = 0xff;
							length_sum += sizeof(color_table);
							for (int init = 0; init < 256; ++init)
							{
								int index = init * 8;
								color_table.entries[index + 2] = init;
								//color_table.entries[index + 4] = init;
								//color_table.entries[index + 6] = init;
							}
						}

						mmquicktime_time_to_sample_table time_to_sample_table;
						memset(&time_to_sample_table, 0, sizeof(time_to_sample_table));
						big_endian(24, time_to_sample_table.atom_size);
						memcpy(time_to_sample_table.atom_name, "stts", 4);
						time_to_sample_table.entry_count[3] = 1;
						length_sum += 16;
	
						// Sample durations array
						unsigned char sample_durations[8];
						big_endian(frame_count, sample_durations);
						big_endian(time_per_frame, &sample_durations[4]);
						length_sum += 8;

						/**
						mmquicktime_keyframe_table keyframe_table; // sync sample atom
						memset(&keyframe_table, 0, sizeof(keyframe_table));
						memcpy(keyframe_table.atom_name, "stss", 4);
						keyframe_table.entry_count[3] = 1;
						keyframe_table.atom_size[3] = 20;

						char keyframe_list[4] = { 0, 0, 0, 1 };
						length_sum += 20;
						**/

						mmquicktime_chunk_table chunk_table; // AKA Sample-to-chunk
						memset(&chunk_table, 0, sizeof(chunk_table));
						memcpy(chunk_table.atom_name, "stsc", 4);
						chunk_table.entry_count[3] = 1;
						chunk_table.atom_size[3] = 28;

						char chunk_list[12] = { 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1 };

						mmquicktime_sample_size_table sample_size_table;
						memset(&sample_size_table, 0, sizeof(sample_size_table));
						memcpy(sample_size_table.atom_name, "stsz", 4);
						big_endian(20, sample_size_table.atom_size);
						big_endian(frame_size, sample_size_table.sample_size);
						length_sum += 20;
	
						/***
						unsigned char sample_size_entries[frame_count * 4];
						for (int init = 0; init < frame_count; ++init)
							big_endian(frame_width * frame_height * colors, 
								&sample_size_entries[init * 4]);
						length_sum += (frame_count * 4);
						big_endian(20 + (frame_count * 4), sample_size_table.atom_size);
						big_endian(frame_count, sample_size_table.entry_count);
						***/

						mmquicktime_sample_offset_table sample_offset_table;
						memset(&sample_offset_table, 0, sizeof(sample_offset_table));
						memcpy(sample_offset_table.atom_name, "stco", 4);
						big_endian(frame_count, sample_offset_table.entry_count);

						unsigned char sample_offsets[4 * frame_count];
						for (int frame = 0, offset = 8; frame < frame_count; 
								++frame, offset += frame_size)
							big_endian(offset, &sample_offsets[frame << 2]);
						
						int offset_table_length = 16 + (4 * frame_count);
						big_endian(offset_table_length, sample_offset_table.atom_size);
						length_sum += offset_table_length;

					big_endian(length_sum - sample_table_start, sample_table.atom_size);

				big_endian(length_sum - media_information_container_start,
					media_information_container.atom_size);

			big_endian(length_sum - media_container_start, media_container.atom_size);

		big_endian(length_sum - track_container_start, track_container.atom_size);

	big_endian(length_sum, movie_container.atom_size);

	// Write to file
	write(descriptor, &movie_container, 8);
	write(descriptor, &movie_header, little_endian(movie_header.atom_size));
	write(descriptor, &track_container, 8);
	write(descriptor, &track_header, little_endian(track_header.atom_size));
	write(descriptor, &media_container, 8);
	write(descriptor, &media_header, little_endian(media_header.atom_size));
	write(descriptor, &video_media_handler, sizeof(video_media_handler));
	// write(descriptor, video_handler_name, strlen(video_handler_name) + 1);
	write(descriptor, &media_information_container, 8);
	write(descriptor, &video_media_information, little_endian(video_media_information.atom_size));
	write(descriptor, &video_alias_handler, sizeof(video_alias_handler));
	// write(descriptor, alias_handler_name, strlen(alias_handler_name) + 1);
	write(descriptor, &data_reference_container, 8);
	write(descriptor, &data_reference, 16);
	write(descriptor, &data_reference_entry, little_endian(data_reference_entry.atom_size));
	write(descriptor, &sample_table, 8);
	write(descriptor, &sample_description, sizeof(sample_description));
	write(descriptor, &sample_description_entry, sizeof(sample_description_entry));
	if (color_mode == MMQUICKTIME_COLOR_MODE_GRAY)
		write(descriptor, &color_table, sizeof(color_table));
	write(descriptor, &time_to_sample_table, 16);
	write(descriptor, sample_durations, 8);
	//write(descriptor, &keyframe_table, 16);
	//write(descriptor, keyframe_list, 4);
	write(descriptor, &chunk_table, 16);
	write(descriptor, chunk_list, 12);
	write(descriptor, &sample_size_table, 20);
	write(descriptor, &sample_offset_table, 16);
	write(descriptor, sample_offsets, 4 * frame_count);

	// Write movie data length
	unsigned char big_endian_length[4];
	big_endian(movie_data_length, big_endian_length);
	lseek(descriptor, 0, SEEK_SET);
	write(descriptor, big_endian_length, 4);

	// Close the file and we're done
	close(descriptor);
	descriptor = -1;

	return length_sum;
};

