/***************************************************************************
	mmsuper8deflicker - Adjusts luminance midpoint of series of frames 
		so each has the same average overall luminance

			-------------------
        begin           : 5/31/2004, 1/15/2005
        copyright       : (C) 2004 by Michael Minn
        email           : see www.michaelminn.com
                        -------------------

***************************************************************************/

#include <stdlib.h>
#include "mmpnm.h"

int print_syntax()
{
	fprintf(stderr, "mmsuper8deflicker - Adjusts luminance midpoint so each PNM/PGM/PPM frame\n");
	fprintf(stderr, "    in a series of frames has the same average overall luminance\n");
	fprintf(stderr, "mmsuper8deflicker [-l red green blue gray] frame_file ...\n\n");
	fprintf(stderr, "	-l specify desired luminance values (do not compute an average from all frames)\n");
	return 0;
}

int main(int argc, char **argv)
{
	int average[4] = {0,0,0,0};
	int frame_count = 0;

	fprintf(stderr, "mmsuper8deflicker %s\n", VERSION);

	for (int argument = 1; argument < argc; ++argument)
		if (!strcmp(argv[argument], "-l") && ((argument + 4) < argc))
		{
			average[0] = atoi(argv[argument + 1]);
			average[1] = atoi(argv[argument + 2]);
			average[2] = atoi(argv[argument + 3]);
			average[3] = atoi(argv[argument + 4]);
			argument = argc;
		}

		else if (!strstr(argv[argument], ".ppm") 
			 && !strstr(argv[argument], ".pgm")
			 && !strstr(argv[argument], ".pnm"))
			return print_syntax();
		else
		{
			mmpnm_file file;
			if (file.load(argv[argument]) < 0)
				continue;

			else if (file.width < file.height)
				fprintf(stderr, "File %s does not seem to be a single frame image\n", argv[argument]);

			else
			{
				int luminance[4];
				for (int color = 0; color < 4; ++color)
				{
					luminance[color] = file.find_average_luminance(color);
					average[color] += luminance[color];
				}
				fprintf(stderr, "%s: %d x %d (luminance %d %d %d %d)\n", argv[argument], file.width, file.height,
					luminance[0], luminance[1], luminance[2], luminance[3]);
				++frame_count;
			}
		}

	for (int color = 0; (color < 4) && frame_count; ++color)
		average[color] /= frame_count;

	fprintf(stderr, "Average luminance: %d %d %d %d\n", average[0], average[1], average[2], average[3]);

	frame_count = 0;
	for (int argument = 1; argument < argc; ++argument)
	{
		mmpnm_file file;
		if (file.load(argv[argument]) < 0)
			continue;

		else if (file.width < file.height)
			fprintf(stderr, "File %s does not seem to be a single frame image\n", argv[argument]);

		else
		{
			int midpoint[4] = {127, 127, 127, 127};
			for (int color = 0; color < 4; ++color)
			{
				int luminance = file.find_average_luminance(color);
				while ((luminance < average[color]) && (midpoint[color] < 254))
				{
					mmpnm_file work;
					work.copy(&file);
					--midpoint[color];
					work.adjust_midpoint(color, midpoint[color]);
					luminance = work.find_average_luminance(color);
printf("	Midpoint %d for color %d = luminance %d\n", midpoint[color], color, luminance);
				}
				while ((luminance > average[color]) && (midpoint[color] > 1))
				{
					mmpnm_file work;
					work.copy(&file);
					++midpoint[color];
					work.adjust_midpoint(color, midpoint[color]);
					luminance = work.find_average_luminance(color);
printf("	Midpoint %d for color %d = luminance %d\n", midpoint[color], color, luminance);
				}
				file.adjust_midpoint(color, midpoint[color]);
			}

			fprintf(stderr, "%s: Adjusted midpoints %d %d %d %d ==========================\n", argv[argument],
				midpoint[0], midpoint[1], midpoint[2], midpoint[3]);
		
			file.save(argv[argument]);
			++frame_count;
		}
	}

	fprintf(stderr, "%d frames processed\n", frame_count);

	return 0;
}
