/***************************************************************************
	mmsuper8jitter.cpp - Adds jitter and grain noise 

			-------------------
        begin           : 7/6/2004
        copyright       : (C) 2004 by Michael Minn
        email           : see www.michaelminn.com
                        -------------------

    This program is free software; you can redistribute it and/or modify
    it under the terms of version 2 of the GNU General Public License as 
    published by the Free Software Foundation.

***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "mmpnm.h"

#define MMSUPER8_DEFAULT_JITTER		1
#define MMSUPER8_DEFAULT_FLICKER	16
#define MMSUPER8_DEFAULT_GRAIN_LEVEL	12

int print_syntax()
{
	fprintf(stderr, "SYNTAX: mmsuper8jitter [options] [pnmfile1 pnmfile2 ...]\n"
			"  Options:\n"
			"    -jitter <jitter_pixels>\n"
			"    -flicker <flicker_level>\n"
			"    -grainlevel <grain_level>\n"
			"    -graincount <grain_count>\n");
	return 0;
}

int main(int argc, char **argv)
{
	int jitter = MMSUPER8_DEFAULT_JITTER;
	int flicker = MMSUPER8_DEFAULT_FLICKER;
	int grain_level = MMSUPER8_DEFAULT_GRAIN_LEVEL;
	int grain_count = -1;

	fprintf(stderr, "%s\n", VERSION);

	if (argc <= 1)
		return print_syntax();

	for (int argument = 1; argument < argc; ++argument)
		if (strstr(argv[argument], "-flicker") && ((argument + 1) < argc))
		{
			++argument;
			flicker = atoi(argv[argument]);
		}

		else if (strstr(argv[argument], "-jitter") && ((argument + 1) < argc))
		{
			++argument;
			jitter = atoi(argv[argument]);
		}

		else if (strstr(argv[argument], "-graincount") && ((argument + 1) < argc))
		{
			++argument;
			grain_count = atoi(argv[argument]);
		}

		else if (strstr(argv[argument], "-grainlevel") && ((argument + 1) < argc))
		{
			++argument;
			grain_level = atoi(argv[argument]);
		}

		else if (!strstr(argv[argument], ".pnm") && !strstr(argv[argument], ".pgm"))
		{
			fprintf(stderr, "Invalid argument: %s\n", argv[argument]);
			return print_syntax();
		}

	int frame_count = 0;
	for (int argument = 1; argument < argc; ++argument)
		if (strstr(argv[argument], ".pnm") || strstr(argv[argument], ".pgm"))
		{
			mmpnm_file file;
			if (file.load(argv[argument]) >= 0)
			{
				fprintf(stderr, "%s: %d x %d\n", argv[argument], 
					file.width, file.height);

				file.add_jitter(jitter);
				file.add_grain(grain_level, grain_count);
				file.add_flicker(flicker);
				file.save(argv[argument]);
			}
		}

	fprintf(stderr, "%d frames saved\n", frame_count);

	return 0;
}
