/***************************************************************************
	mmsuper8avi.cpp - Converts PNM super8 frame scans to a .AVI file

			-------------------
        begin           : 5/31/2004, 1/15/2005
        copyright       : (C) 2004 by Michael Minn
        email           : see www.michaelminn.com
                        -------------------

***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include "mmpnm.h"
//#include "mmavi.h"
#include "mmquicktime.h"


/************************************************************************
		Command Line Interface
************************************************************************/

int print_syntax()
{
	fprintf(stderr, "mmsuper8mov - Combines .ppm/.pgm/.pnm frames into a single uncompressed .mov (Quicktime) file\n\n");

	fprintf(stderr, "mmsuper8mov [options] framefile...\n\n");

	fprintf(stderr, "	-o output_file \n");
	fprintf(stderr, "		Specifies name of output file (default is movie.mov)\n\n");
			
	fprintf(stderr, "	-b\n");
	fprintf(stderr, "		Output file should be black-and-white (i.e. grayscale)\n\n");
			
	fprintf(stderr, "	-rate fps\n");
	fprintf(stderr, "		Specifies frame rate (default is 18 bps)\n");

	return 0;
}

int main(int argc, char **argv)
{
	int frames_per_second = 18;
	int color_mode = MMQUICKTIME_COLOR_MODE_RGB;
	char *output_filename = "movie.mov";

	fprintf(stderr, "%s\n", VERSION);

	// First pass through arguments for options
	for (int argument = 1; argument < argc; ++argument)
		if (strstr(argv[argument], "-o") && ((argument + 1) < argc))
		{
			++argument;
			output_filename = argv[argument];
		}

		else if (strstr(argv[argument], "-rate") && ((argument + 1) < argc))
		{
			++argument;
			frames_per_second = atoi(argv[argument]);
		}

		else if (strcmp(argv[argument], "-b") == 0)
		{
			color_mode = MMQUICKTIME_COLOR_MODE_GRAY;
			fprintf(stderr, "Output will be black and white (grayscale)\n");
		}

		else if (strncmp(argv[argument], "-h", 2) == 0)
			return print_syntax();

		else if (strstr(argv[argument], ".ppm") 
			 || strstr(argv[argument], ".pgm")
			 || strstr(argv[argument], ".pnm"))
			continue; // processed below in 2nd pass

		else
		{
			fprintf(stderr, "Invalid argument: %s\n", argv[argument]);
			return print_syntax();
		}

	// Open output file
	mmquicktime_file quicktime_file;
	if (quicktime_file.open_file(output_filename, color_mode, frames_per_second) < 0)
		return -1;

	// Second pass through arguments to process input files and context sensitive parameters
	int frame_count = 0;
	for (int argument = 1; argument < argc; ++argument)
		if (argv[argument][0] != '-')
		{
			mmpnm_file file;
			if (file.load(argv[argument]) >= 0)
			{
				fprintf(stderr, "%s: %d x %d\n", argv[argument], file.width, file.height);

				if (file.width < file.height)
					fprintf(stderr, "  File does not appear to be a frame file\n");
				else
				{
					quicktime_file.add_frame(&file);
					++frame_count;
				}
			}
		}

	quicktime_file.close_file();

	fprintf(stderr, "%d frames saved\n", frame_count);

	return 0;
}
