/***************************************************************************
	mmsuper8normalize - Normalizes colors in .ppm/.pgm/.pnm frame files

			-------------------
        begin           : 5/31/2004, 1/15/2005
        copyright       : (C) 2004 by Michael Minn
        email           : see www.michaelminn.com
                        -------------------

***************************************************************************/

#include <stdlib.h>
#include "mmpnm.h"

int print_syntax()
{
	fprintf(stderr, "mmsuper8normalize - Normalizes colors in .ppm/.pgm/.pnm frame files\n\n");

	fprintf(stderr, "mmsuper8normalize [options] frame_file ...\n\n");

	fprintf(stderr, "	-red min max\n");
	fprintf(stderr, "		Overrides automatic detection of min/max red color values\n\n");

	fprintf(stderr, "	-green min max\n");
	fprintf(stderr, "		Overrides automatic detection of min/max red color values\n\n");

	fprintf(stderr, "	-blue min max\n");
	fprintf(stderr, "		Overrides automatic detection of min/max red color values\n\n");

	fprintf(stderr, "	-mid midpoint\n");
	fprintf(stderr, "		Specifies midpoint for adjusting luminance after normalization\n\n");

	return 0;
}

int main(int argc, char **argv)
{
	int normalize = 0;
	int midpoint = -1;
	int color_high[3] = { 255, 255, 255 };
	int color_low[3] = { 0, 0, 0 };

	fprintf(stderr, "mmsuper8normalize %s\n", VERSION);

	int frame_count = 0;
	for (int argument = 1; argument < argc; ++argument)
		if (strncmp(argv[argument], "-n", 2) == 0)
			normalize = 1;

		else if (strstr(argv[argument], "-red") && ((argument + 2) < argc))
		{
			color_low[0] = atoi(argv[argument + 1]);
			color_high[0] = atoi(argv[argument + 2]);
			argument += 2;
			fprintf(stderr, "Expanding red between %d and %d\n", color_low[0], color_high[0]);
		}	

		else if (strstr(argv[argument], "-green") && ((argument + 2) < argc))
		{
			color_low[1] = atoi(argv[argument + 1]);
			color_high[1] = atoi(argv[argument + 2]);
			argument += 2;
			fprintf(stderr, "Expanding green between %d and %d\n", color_low[1], color_high[1]);
		}	

		else if (strstr(argv[argument], "-blue") && ((argument + 2) < argc))
		{
			color_low[2] = atoi(argv[argument + 1]);
			color_high[2] = atoi(argv[argument + 2]);
			argument += 2;
			fprintf(stderr, "Expanding blue between %d and %d\n", color_low[2], color_high[2]);
		}	

		else if (strstr(argv[argument], "-mid") && ((argument + 1) < argc))
		{
			midpoint = atoi(argv[argument + 1]);
			argument += 1;
			fprintf(stderr, "Setting luminance midpoint to %d\n", midpoint);
		}	

		else if (strstr(argv[argument], ".ppm") 
			 || strstr(argv[argument], ".pgm")
			 || strstr(argv[argument], ".pnm"))
		{
			mmpnm_file file;
			if (file.load(argv[argument]) < 0)
				continue;

			else if (file.width < file.height)
				fprintf(stderr, "  File does not seem to be a single frame image\n");

			else
			{
				fprintf(stderr, "%s: %d x %d\n", argv[argument], file.width, file.height);

				file.normalize_gray();

				for (int color = 0; color < 3; ++color)
					if ((color_low[color] > 0) || (color_high[color] < 255))
						file.adjust_color(color, color_low[color], color_high[color]);
					else
						file.normalize_color(color);

				if (midpoint > 0)
					for (int color = 0; color < 4; ++color)
						file.adjust_midpoint(color, midpoint);
		
				file.save(argv[argument]);
				++frame_count;
			}
		}

		else
		{
			fprintf(stderr, "Invalid argument: %s\n", argv[argument]);
			return print_syntax();
		}

	fprintf(stderr, "%d frames processed\n", frame_count);

	return 0;
}
