/***************************************************************************
	mmsuper8play - Player for PNM frame files

			-------------------
        begin           : 6/7/2004
        copyright       : (C) 2004 by Michael Minn
        email           : see www.michaelminn.com
                        -------------------

    This program is free software; you can redistribute it and/or modify
    it under the terms of version 2 of the GNU General Public License as 
    published by the Free Software Foundation.

***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <math.h>
#include "mmpnm.h"

#include <X11/X.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h> // includes keysymdef.h - needed for XK_* constants

#define MMSUPER8PLAY_VERSION "MMSUPER8PLAY v2004.06.26"
#define MMSUPER8PLAY_SCREEN_HEIGHT 400
#define MMSUPER8PLAY_SCREEN_WIDTH 540
#define MMSUPER8PLAY_DEFAULT_FPS 18
#define MMSUPER8PLAY_REFRESH_MS 20

class mmsuper8play
{
	public:
	mmsuper8play(int argc, char **argv);
	int refresh_screen();
	int list_screen_info();
	int change_pitch(int new_pitch);

	int y_shift;
	int file_count;
	int current_fps;
	int current_file;
	char **file_list;
	struct timeval last_frame_time;
	XtAppContext application;
	Widget shell;
	Pixmap pixmap;
};

/**************************** Callback Functions ****************************/

void refresh_callback(XtPointer context_pointer, XtIntervalId *interval)
{
	mmsuper8play *player = (mmsuper8play*) context_pointer;
	
	player->refresh_screen();

	//if (player->current_fps == 0)
	//	XtAppAddTimeOut(player->application, 1000, refresh_callback, context_pointer);
	//else
		XtAppAddTimeOut(player->application, MMSUPER8PLAY_REFRESH_MS, 
				refresh_callback, context_pointer);
}

void key_callback(Widget widget, XtPointer client_data, XEvent *event, Boolean *continue_flag)
{
        // KeySym definitions are in /usr/X11R6/include/X11/keysymdef.h
	int keysym = XKeycodeToKeysym(event->xkey.display, event->xkey.keycode, 0);
	mmsuper8play *player = (mmsuper8play*) client_data;
	// printf("Key press %x %x\n", event->type, keysym);

	if ((event->type & KeyPressMask) && (keysym == XK_Left))
	{
		player->current_fps -= 2;
		printf("%d fps\n", player->current_fps);
	}
		
	else if ((event->type & KeyPressMask) && (keysym == XK_Right))
	{
		player->current_fps += 2;
		printf("%d fps\n", player->current_fps);
	}
	
	else if ((event->type & KeyPressMask) && (keysym == XK_Up))
	{
		++player->y_shift;
		printf("Shift %d pixels\n", -player->y_shift);
	}

	else if ((event->type & KeyPressMask) && (keysym == XK_Down))
	{
		--player->y_shift;
		printf("Shift %d pixels\n", -player->y_shift);
	}
	
	else if ((event->type & KeyPressMask) && (keysym == XK_Home))
	{
		player->current_fps = 0;
		player->current_file = 1;
	}

	else if ((event->type & KeyPressMask) && (keysym == XK_End))
	{
		player->current_fps = 0;
		player->current_file = player->file_count - 1;
	}

	else if ((event->type & KeyPressMask) && (keysym == XK_space))
	{
		player->current_fps = 0;
		if (player->current_file < (player->file_count - 1))
			++player->current_file;
	}

	else if ((event->type & KeyPressMask) && (keysym == XK_Page_Up))
	{
		if ((player->current_file + MMSUPER8PLAY_DEFAULT_FPS) < player->file_count)
			player->current_file += MMSUPER8PLAY_DEFAULT_FPS;
	}

	else if ((event->type & KeyPressMask) && (keysym == XK_Page_Down))
	{
		if ((player->current_file - MMSUPER8PLAY_DEFAULT_FPS) > 0)
			player->current_file -= MMSUPER8PLAY_DEFAULT_FPS;
	}

	else if ((event->type & KeyPressMask) && (keysym == XK_Return))
	{
		if (player->current_fps == 0)
			player->current_fps += MMSUPER8PLAY_DEFAULT_FPS;
		else
			player->current_fps = 0;
	}

	/**
	else if ((event->type & KeyPressMask) && (player->current_fps == 0) && (player->y_shift != 0)
			&& ((keysym == XK_S) || (keysym == XK_s)))
	{
		printf("Shifting %s by %d pixels\n", 
			player->file_list[player->current_file], player->y_shift);
				
		mmpnm_file input_file;
		if (input_file.load(player->file_list[player->current_file]) < 0)
			return;
		else if (input_file.shift(0, player->y_shift) < 0)
			return;
		else
			input_file.save(player->file_list[player->current_file]);
	}
	**/
		
	else if ((event->type & KeyPressMask) && ((keysym == XK_Q) || (keysym == XK_q)))
		exit(0);
}



/**************************** Classed Functions ****************************/
			
int mmsuper8play::refresh_screen()
{
	struct timeval current_time;
	gettimeofday(&current_time, NULL);
	int ms_since_last_frame = ((current_time.tv_sec - last_frame_time.tv_sec) * 1000)
		+ ((current_time.tv_usec - last_frame_time.tv_usec) / 1000);

	//printf("%d %d\n", ms_since_last_frame, abs(1000 / current_fps));

	if ((current_fps != 0) && (ms_since_last_frame >= abs(1000 / current_fps)))
	{
		last_frame_time = current_time;
		if ((current_fps > 0) && (current_file < (file_count - 1)))
			++current_file;
		else if ((current_fps < 0) && (current_file > 1))
			--current_file;
	}

	// printf("%d %d\n", current_file, ms_since_last_frame);
	if (current_file < 1)
		return 0;

	mmpnm_file input_file;
	int status = input_file.load(file_list[current_file]);
	if (status < 0)
		return status;

	XtVaSetValues(shell, XtNtitle, file_list[current_file], NULL);
	
	Display *display = XtDisplay(shell);
	Window window = XtWindow(shell);
	Visual *visual = DefaultVisual(display, DefaultScreen(display));
	GC graphic_context = DefaultGC(display, DefaultScreen(display));
	//XSetWindowBackground(XtDisplay(shell),window, BlackPixelOfScreen(DefaultScreenOfDisplay(display)));
		
	int depth = DefaultDepth(display, DefaultScreen(display));
	int bytes_per_pixel = depth / 8;

	if (bytes_per_pixel > 4)
	{
		fprintf(stderr, "No support provided for graphic displays over 32 bit depth\n");
		return -EINVAL;
	}

	int masks[3] = { visual->blue_mask, visual->green_mask, visual->red_mask };
	int shift[3] = { 0, 0, 0 };
	for (int color = 0; color < 3; ++color)
	{
		for (int scan = masks[color]; scan; scan >>= 1)
			++shift[color];

		shift[color] = (8 * (color + 1)) - shift[color];
	}
	
	//printf("Shifts %d %x, %d %x, %d %x\n", shift[2], masks[2],
	//	shift[1], masks[1], shift[0], masks[0]);

	int data_length = input_file.width * input_file.height * bytes_per_pixel;
	char *pixmap_data = XtMalloc(data_length);
	memset(pixmap_data, 0, data_length);

	mmpnm_file shifted;
	shifted.copy(&input_file);
	shifted.shift(0, y_shift);
	
	for (int y = 0, source = 0, destination = 0; y < input_file.height; ++y)
		for (int x = 0; x < input_file.width; ++x, source += 3, destination += bytes_per_pixel)
		{
			int total = 0;
			int red = 0;
			int green = 0;
			int blue = 0;
			if ((y >= 0) && (y < input_file.height))
			{
				red = shifted.color_data[source];
				green = shifted.color_data[source + 1];
				blue = shifted.color_data[source + 2];
			}

			total |= ((blue >> shift[0]) & masks[0]);
			total |= (((green << 8) >> shift[1]) & masks[1]);
			total |= (((red << 16) >> shift[2]) & masks[2]); 
			memcpy(&pixmap_data[destination], &total, bytes_per_pixel);
		}


	XImage *image = XCreateImage(display, visual, depth, ZPixmap, 0,
				pixmap_data, 
				input_file.width, input_file.height, 8,
				input_file.width * bytes_per_pixel);
	
	if (!image)
	{
		perror("Failure creating image");
		return -1;
	}
	
	XPutImage(display, window, graphic_context, image, 0, 0, 0, 0,
		MMSUPER8PLAY_SCREEN_WIDTH, MMSUPER8PLAY_SCREEN_HEIGHT);

	// Note that this deletes the "image" structure as well as the pixmap_data
	XDestroyImage(image);

	return 0;
}

int mmsuper8play::list_screen_info()
{
	Screen *screen = XtScreen(shell);
	Visual *visual = screen->root_visual;
	
	printf("%d Available depths\n", screen->ndepths);
	for (int depth = 0; depth < screen->ndepths; ++depth)
		for (int visual = 0; visual < screen->depths[depth].nvisuals; ++visual)
			printf("  Depth %d, Visual %d: %d bits per color, masks %x/%x/%x\n",
				depth, visual,
				(int) screen->depths[depth].visuals[visual].bits_per_rgb,
				(int) screen->depths[depth].visuals[visual].red_mask,
				(int) screen->depths[depth].visuals[visual].green_mask,
				(int) screen->depths[depth].visuals[visual].blue_mask);

	printf("Root Visual\n");
	printf("  %d bits per pixel\n", screen->root_depth);
	printf("  Masks %x %x %x, bits per color %d\n", 
		(int) visual->red_mask, (int) visual->green_mask, 
		(int) visual->blue_mask, (int) visual->bits_per_rgb);

	return 0;
}

mmsuper8play::mmsuper8play(int argc, char **argv)
{
	memset(this, 0, sizeof(*this));

	String default_resources[] = {
		"mmsuper8play*background: black",
		"mmsuper8play*foreground: black",
		NULL };

	file_count = argc;
	file_list = argv;
	current_file = 1;

	shell = XtVaOpenApplication (&application, "player", NULL, 0,
		&argc, argv, default_resources, sessionShellWidgetClass,
		XtNwidth, MMSUPER8PLAY_SCREEN_WIDTH,
		XtNheight, MMSUPER8PLAY_SCREEN_HEIGHT,
		XtNtitle, "MMSUPER8PLAY",
		//XmNmwmDecorations, MWM_DECOR_ALL | MWM_DECOR_RESIZEH,
		NULL);
	
	XtAddEventHandler(shell, 0xffff, 0, (XtEventHandler) &key_callback, (XtPointer) this);
	
	XtRealizeWidget(shell);
	
	gettimeofday(&last_frame_time, NULL);
	current_fps = MMSUPER8PLAY_DEFAULT_FPS;
}

int main(int argc, char **argv)
{
	fprintf(stderr, "%s\n", MMSUPER8PLAY_VERSION);

	mmsuper8play player(argc, argv);

	player.list_screen_info();

	XtAppAddTimeOut(player.application, MMSUPER8PLAY_REFRESH_MS, refresh_callback, (XtPointer) &player);

	XtAppMainLoop(player.application);

	return 0;
}

